# Would be nice to have a class for this, that automatically switches between cuda and cpu


def simple_moving_average(df, window):
    res = df["close"].rolling(window).mean()
    res.name = f"sma_{window}"
    return res


def exponential_moving_average(df, window):
    # had to adjust for cuda
    cum_sum = df["close"].rolling(window).sum()
    cum_count = df["close"].rolling(window).count()
    ema = cum_sum / cum_count
    ema.name = f"ema_{window}"
    return ema


def bollinger_bands(df, window):
    sma = simple_moving_average(df, window)
    std = df["close"].rolling(window).std()
    upper = sma + 2 * std
    lower = sma - 2 * std
    upper.name = f"bbupper_{window}"
    lower.name = f"bblower_{window}"
    return upper, lower


def rsi(df, window):
    delta = df["close"].diff()
    up_days = delta.copy()
    up_days[delta <= 0] = 0.0
    down_days = abs(delta.copy())
    down_days[delta > 0] = 0.0
    RS_up = up_days.rolling(window).mean()
    RS_down = down_days.rolling(window).mean()
    rsi = 100 - 100 / (1 + RS_up / RS_down)
    rsi.name = f"rsi_{window}"
    return rsi


def macd(df, window_fast, window_slow):
    ema_fast = exponential_moving_average(df, window_fast)
    ema_slow = exponential_moving_average(df, window_slow)
    macd = ema_fast - ema_slow
    macd.name = f"macd_{window_fast}_{window_slow}"
    return macd


# ATR
def average_true_range(df, window):
    tr = df[["high", "low", "close"]].max(axis=1) - df[["high", "low", "close"]].min(
        axis=1
    )
    atr = tr.rolling(window).mean()
    atr.name = f"atr_{window}"
    return atr



from signalslite.data_downloader import update_daily_data
from signalslite.primary_features import generate_primary_features
from signalslite.secondary_features import update_secondary_features
from signalslite.scale_data import scale_data
from signalslite.merge_targets import prepare_merged_data
from signalslite.constants import Directories
from credentials import EODHD_API_KEY

if __name__ == "__main__":
    dir_config = Directories()
    dir_config.set_data_dir("data")

    EODHD_API_KEY = None

    print(dir_config)
    update_daily_data(dir_config.DATA_DIR, dir_config.DAILY_DATA_DIR, EODHD_API_KEY)
    generate_primary_features(dir_config)
    update_secondary_features(dir_config)
    scale_data(dir_config, FROM_SCRATCH=False)
    prepare_merged_data(dir_config)



import pandas as pd
import os
from pathlib import Path
import gc
import multiprocessing
from joblib import Parallel, delayed
from tqdm import tqdm

import cudf
import numba
from numba import cuda
import numpy as np
import numerapi

# parallelize the process on all columns using joblib
import multiprocessing
from joblib import Parallel, delayed
from tqdm import tqdm

from signalslite.data_utils import (
    load_recent_data_from_file,
    save_daily_data,
    save_in_folders,
    get_latest_date,
    read_available_dates,
)
from signalslite.constants import Directories


def update_historical_file(dir_config):
    napi = numerapi.SignalsAPI(verbosity="info")
    # get the latest date
    pd.read_csv(napi.HISTORICAL_DATA_URL).to_csv(
        dir_config.DATA_DIR / "numerai_signals_historical.csv", index=False
    )


def merge_data(historical_df: pd.DataFrame, dir_config: Directories):
    dates = read_available_dates(dir_config.DAILY_SCALED_FEATURES_DIR)

    merged_df = []

    for i in tqdm(range(0, len(dates), 200)):
        _tmp = load_recent_data_from_file(
            dir_config.DAILY_SCALED_FEATURES_DIR, n_days=200, ascending=True, offset=i
        )
        _tmp = _tmp.reset_index(drop=True)
        _tmp = _tmp.sort_values(["date", "bloomberg_ticker"])
        feature_columns = [f for f in _tmp.columns if f.startswith("feature")]

        _historical_dates = historical_df["date"].unique()
        # find common dates
        common_dates = list(
            set(_tmp["date"].unique()).intersection(set(_historical_dates))
        )

        if len(common_dates) == 0:
            continue

        # merge historical_df and _tmp on date and bloomberg_ticker
        _merged = pd.merge(
            historical_df,
            _tmp,
            how="right",
            left_on=["date", "bloomberg_ticker"],
            right_on=["date", "bloomberg_ticker"],
        )
        _merged = _merged.dropna(subset=["target_20d"], axis=0)

        # assert abs(_merged["feature_2_ratio_rsi_50_close"].mean() - 2) < 0.05

        if len(_merged) > 0:
            merged_df.append(_merged)

    merged_df = pd.concat(merged_df)

    return merged_df


def prepare_live(dir_config):
    latest_date = get_latest_date(dir_config.DAILY_SCALED_FEATURES_DIR)
    print("latest date: ", latest_date)
    # load the latest data
    df = load_recent_data_from_file(
        dir_config.DAILY_SCALED_FEATURES_DIR, n_days=1, ascending=False
    )
    df = df.reset_index(drop=True)
    print("df shape: ", df.shape)

    return df


def prepare_merged_data(dir_config):
    update_historical_file(dir_config)
    historical_df = pd.read_csv(dir_config.DATA_DIR / "numerai_signals_historical.csv")
    historical_df["date"] = pd.to_datetime(
        historical_df["friday_date"], format="%Y%m%d"
    )

    merged_df = merge_data(historical_df, dir_config)
    live_df = prepare_live(dir_config)

    merged_df.to_parquet(
        dir_config.DATA_DIR / "merged_data_historical.parquet", index=False
    )
    live_df.to_parquet(dir_config.DATA_DIR / "merged_data_live.parquet", index=False)


if __name__ == "__main__":
    dir_config = Directories()
    dir_config.set_data_dir("data")

    print(dir_config)
    prepare_merged_data(dir_config)



import pandas as pd
import numpy as np
import os
from pathlib import Path
import gc
import multiprocessing
from joblib import Parallel, delayed
from tqdm import tqdm

from signalslite.data_utils import (
    load_recent_data_from_file,
    save_in_folders,
    get_latest_date,
    read_available_dates,
)
from signalslite.constants import Directories

from signalslite.technical_features import (
    simple_moving_average,
    exponential_moving_average,
    bollinger_bands,
    rsi,
    macd,
    average_true_range,
)

try:
    import cudf
    import numba
    from numba import cuda

    USE_CUDF = True
except ImportError:
    USE_CUDF = False
    print("cudf not found, using pandas")


def load_recent_data(DAILY_DATA_DIR, n_days):
    recent_data = (
        load_recent_data_from_file(DAILY_DATA_DIR, n_days=n_days, ascending=False)
        .reset_index()
        .sort_values(by=["bloomberg_ticker", "date"])
    )

    cols = recent_data.select_dtypes(include=["float32", "int64"]).columns

    recent_data[cols] = recent_data[cols].astype("float16")

    # filter out tickers with less than 100 days of data
    recent_data = recent_data.groupby("bloomberg_ticker").filter(lambda x: len(x) > 100)
    recent_data = recent_data.groupby("date").filter(lambda x: len(x) > 500)
    gc.collect()

    recent_data.info()

    return recent_data


def compute_features(df, function_to_window, use_cudf:bool):
    features = []
    for func, windows in function_to_window.items():
        for window in windows:
            # pass windows as a tuple if the function takes more than one window
            if isinstance(window, tuple):
                _feat = func(df, *window)
            else:
                _feat = func(df, window)

            if isinstance(_feat, tuple):
                features.extend(_feat)
            else:
                features.append(_feat)

    # print type of features
    if use_cudf:
        print("features type: cudf")
        cated = cudf.concat(features, axis=1).astype("float32").add_prefix("feature_1_")
    else:
        print("features type: pandas")
        cated = pd.concat(features, axis=1).astype("float32").add_prefix("feature_1_")
    return cated


def generate_features(recent_data, function_to_window, use_cudf:bool):
    tickers_list = recent_data["bloomberg_ticker"].unique().tolist()

    # iterate over ticker chunks in 500
    res = []
    for i in tqdm(range(0, len(tickers_list), 1000)):
        tickers = tickers_list[i : i + 1000]
        # print(tickers)
        tickers_data = recent_data[recent_data["bloomberg_ticker"].isin(tickers)]
        if use_cudf:
            _df_gpu = cudf.from_pandas(tickers_data)
        else:
            _df_gpu = tickers_data

        _res = compute_features(_df_gpu, function_to_window, use_cudf=use_cudf)
        if use_cudf:
            _res = _res.to_pandas().astype("float16")
            _res["date"] = _df_gpu["date"].to_pandas()
            _res["bloomberg_ticker"] = _df_gpu["bloomberg_ticker"].to_pandas()
            _res["close"] = _df_gpu["close"].to_pandas()
            _res["volume"] = _df_gpu["volume"].to_pandas()
            _res["open"] = _df_gpu["open"].to_pandas()
            _res["high"] = _df_gpu["high"].to_pandas()
            _res["low"] = _df_gpu["low"].to_pandas()
        else:
            _res["date"] = _df_gpu["date"]
            _res["bloomberg_ticker"] = _df_gpu["bloomberg_ticker"]
            _res["close"] = _df_gpu["close"]
            _res["volume"] = _df_gpu["volume"]
            _res["open"] = _df_gpu["open"]
            _res["high"] = _df_gpu["high"]
            _res["low"] = _df_gpu["low"]

        res.append(_res)

        del _df_gpu, _res

        gc.collect()
        res = pd.concat(res, axis=0)
        res = res.dropna(axis=0)

        # convert float 16 to float 32 in a loop
        for col in res.columns:
            if res[col].dtype == "float16":
                res[col] = res[col].astype("float32")
            gc.collect()

        gc.collect()

        return res


def save_features(res, DAILY_PRIMARY_FEATURES_DIR):
    # loop over all unique dates in chunks of 100; save each chunk to a separate file
    res["date_str"] = res["date"].dt.strftime("%Y-%m-%d")
    dates = res["date_str"].unique()

    for i in tqdm(range(0, len(dates), 100)):
        # use save_in_folders function to save each chunk to a separate folder
        _tmp = res[res["date_str"].isin(dates[i : i + 100])]
        save_in_folders(_tmp, DAILY_PRIMARY_FEATURES_DIR)

        del _tmp
        gc.collect()


def generate_primary_features(dir_config):

    try:
        import cudf
        import numba
        from numba import cuda

        USE_CUDF = True
    except ImportError:
        USE_CUDF = False
        print("cudf not found, using pandas")

    n_days_to_load = -1

    # if some of primary features in days are there then take last 1000 days in adjusted data: 1000
    # else take all days in adjusted data: -1
    if os.path.exists(dir_config.DAILY_PRIMARY_FEATURES_DIR):
        raw_data_dates = read_available_dates(dir_config.DAILY_DATA_DIR)
        print(f"raw_data_dates: {len(raw_data_dates)}")
        primary_features_dates = read_available_dates(
            dir_config.DAILY_PRIMARY_FEATURES_DIR
        )

        # difference in days between last date in raw data and last date in primary features
        last_date_raw = np.sort(raw_data_dates)[-1]
        last_date_primary = np.sort(primary_features_dates)[-1]
        diff = (pd.to_datetime(last_date_raw) - pd.to_datetime(last_date_primary)).days
        print(diff)
        print(f"primary_features_dates: {len(primary_features_dates)}")

        n_days_to_load = diff + 1000

    print(f"n_days_to_load: {n_days_to_load}")

    function_to_window: dict = {
        simple_moving_average: [5, 10, 20, 50, 100, 200],
        exponential_moving_average: [5, 10, 20, 50, 100, 200],
        bollinger_bands: [5, 10, 20, 50, 100, 200],
        rsi: [5, 10, 20, 50, 100, 200],
        average_true_range: [5, 10, 20, 50, 100, 200],
        macd: [(12, 26), (20, 50)],
    }

    # load recent data
    recent_data = load_recent_data(dir_config.DAILY_DATA_DIR, n_days_to_load)
    features = generate_features(recent_data, function_to_window, USE_CUDF)
    save_features(features, dir_config.DAILY_PRIMARY_FEATURES_DIR)


if __name__ == "__main__":
    dir_config = Directories()
    generate_primary_features(dir_config)



import os
import dataclasses
from pathlib import Path

# TODO: Load this separately in all files and allow user to change it; potentially for cloud storage facility
# TODO: How about allowing users to define their own class for data_utils.py? (e.g. for cloud storage)
# But first, get this working with the current setup; start on cloud \
# after V0.0.1 in cloud_branch; constraints stays same for efficiency

# use dataclass for constants
class Directories:
    DATA_DIR = Path("data")
    DAILY_DATA_DIR = DATA_DIR / "01_daily_adjusted"
    DAILY_PRIMARY_FEATURES_DIR = DATA_DIR / "02_primary_features"
    DAILY_SECONDARY_FEATURES_DIR = DATA_DIR / "03_secondary_features"
    DAILY_SCALED_FEATURES_DIR = DATA_DIR / "04_scaled_features"

    @classmethod
    def set_data_dir(cls, new_data_dir):
        cls.DATA_DIR = Path(new_data_dir)
        cls.DAILY_DATA_DIR = cls.DATA_DIR / "01_daily_adjusted"
        cls.DAILY_PRIMARY_FEATURES_DIR = cls.DATA_DIR / "02_primary_features"
        cls.DAILY_SECONDARY_FEATURES_DIR = cls.DATA_DIR / "03_secondary_features"
        cls.DAILY_SCALED_FEATURES_DIR = cls.DATA_DIR / "04_scaled_features"



import pandas as pd
import os
from pathlib import Path
import gc
import multiprocessing
from joblib import Parallel, delayed
from tqdm import tqdm

import cudf
import numba
from numba import cuda
import numpy as np

# parallelize the process on all columns using joblib
import multiprocessing
from joblib import Parallel, delayed
from tqdm import tqdm
from typing import List, Tuple, Union, Optional, Dict, Any

from signalslite.data_utils import (
    load_recent_data_from_file,
    save_daily_data,
    save_in_folders,
    get_latest_date,
    read_available_dates,
)
from signalslite.constants import Directories


def apply_cut(df: pd.DataFrame, cols: List[str]) -> pd.DataFrame:
    # NOTE: Parallel is a library for parallelizing operations
    # on a single machine. It is not a distributed framework.
    # The `n_jobs` argument specifies how many CPU cores to use.
    # For example, `n_jobs=10` will use 10 CPU cores.
    _res = Parallel(
        n_jobs=10,
    )(delayed(pd.qcut)(df[col], q=5, labels=False, duplicates="drop") for col in cols)
    _res = pd.concat(_res, axis=1).astype("int8")
    return _res


def apply_cut_cpu(
    recent_data: pd.DataFrame, feature_columns: List[str]
) -> pd.DataFrame:
    recent_data = recent_data.dropna(subset=feature_columns, axis=0)
    recent_data_gpu = cudf.DataFrame.from_pandas(recent_data)

    ranks = (
        recent_data_gpu[["date"] + feature_columns]
        .groupby("date")
        .rank(pct=True, method="first", ascending=True, na_option="keep")
    )
    ranks["close"] = recent_data["close"]
    ranks["date"] = recent_data["date"]
    ranks["bloomberg_ticker"] = recent_data["bloomberg_ticker"]

    ranks_pd = ranks.to_pandas()

    del recent_data_gpu, ranks

    # NOTE: This code will be run on the CPU.
    # We will show you how to run this code on the GPU in the next section.
    res = ranks_pd.groupby("date").apply(lambda df: apply_cut(df, feature_columns))
    res["date"] = ranks_pd["date"]
    res["bloomberg_ticker"] = ranks_pd["bloomberg_ticker"]
    res["close"] = ranks_pd["close"]

    return res


# this code is used to apply a cut on the data, in order to remove some of the data that has been corrupted
# the cut is based on the standard deviation of the data, and is applied on a per-feature basis
# it is applied to the data in chunks of 200 days, in order to avoid memory issues
def scale_data(dir_config, FROM_SCRATCH=False):
    dates = read_available_dates(dir_config.DAILY_SECONDARY_FEATURES_DIR)

    start_index = len(dates) - 1000 if not FROM_SCRATCH else 0

    # iterate over all dates in chunks of 200
    for i in tqdm(range(start_index, len(dates), 200)):
        _tmp = load_recent_data_from_file(
            dir_config.DAILY_SECONDARY_FEATURES_DIR,
            n_days=200,
            ascending=True,
            offset=i,
            dtype="float32",
        )
        _tmp = _tmp.reset_index(drop=True)
        _tmp = _tmp.sort_values(["date", "bloomberg_ticker"])
        _tmp = _tmp.groupby("date").filter(lambda x: len(x) > 10)

        # get all feature columns
        feature_columns = [f for f in _tmp.columns if f.startswith("feature_")]

        # apply cut
        res = apply_cut_cpu(_tmp, feature_columns)

        # save
        save_in_folders(res, dir_config.DAILY_SCALED_FEATURES_DIR)

        del _tmp, res
        gc.collect()


if __name__ == "__main__":
    dir_config = Directories()
    # dir_config.set_data_dir("data")

    FROM_SCRATCH = False

    scale_data(dir_config, FROM_SCRATCH=FROM_SCRATCH)



import os
import gc
import time
from pathlib import Path
import requests
from datetime import datetime
import pandas as pd
from concurrent import futures
from tqdm import tqdm

from joblib import Parallel, delayed
import multiprocessing

from signalslite.constants import Directories

from signalslite.data_utils import (
    load_recent_data_from_file,
    save_daily_data,
    get_latest_date,
    save_in_folders,
)


class YahooDownloaderOHLCV:
    def __init__(self):
        pass

    @staticmethod
    def _download_dividends_yahoo(ticker: str, start_date: str, end_date: str):
        start_epoch = int(start_date.timestamp())
        end_epoch = int(end_date.timestamp())

        dividends = (
            pd.read_csv(
                f"https://query1.finance.yahoo.com/v7/finance/download/{ticker}?period1={start_epoch}&period2={end_epoch}&interval=1d&events=div&includeAdjustedClose=true"
            )
            .dropna()
            .set_index("Date")
        )

        if dividends is not None and len(dividends) > 1:
            dividends["date"] = pd.to_datetime(dividends.index, format="%Y-%m-%d")
            dividends = (
                dividends.reset_index(drop=True)
                .set_index("date")
                .sort_index()
                .rename(columns={"Dividends": "dividend_amount"})
            )

        return dividends

    @staticmethod
    def _download_splits_yahoo(ticker: str, start_date: str, end_date: str):
        start_epoch = int(start_date.timestamp())
        end_epoch = int(end_date.timestamp())

        splits = (
            pd.read_csv(
                f"https://query1.finance.yahoo.com/v7/finance/download/{ticker}?period1={start_epoch}&period2={end_epoch}&interval=1d&events=split&includeAdjustedClose=true"
            )
            .dropna()
            .set_index("Date")
        )


        if splits is not None and len(splits) > 1:
            splits["date"] = pd.to_datetime(splits.index, format="%Y-%m-%d")
            splits = (
                splits.reset_index(drop=True)
                .set_index("date")
                .sort_index()
                .rename(columns={"Stock Splits": "split_factor"})
            )

        return splits

    @classmethod
    def yahoo_download_one(cls, signals_ticker, date_from=None, date_until=None):
        if date_from is None:
            date_from = "2000-01-01"
        if date_until is None:
            date_until = datetime.today().strftime("%Y-%m-%d")

        date_from = datetime.strptime(date_from, "%Y-%m-%d")
        date_until = datetime.strptime(date_until, "%Y-%m-%d")

        start_epoch = int(date_from.timestamp())
        end_epoch = int(date_until.timestamp())

        quotes = None

        quotes = (
            pd.read_csv(
                f"https://query1.finance.yahoo.com/v7/finance/download/{signals_ticker}?period1={start_epoch}&period2={end_epoch}&interval=1d&events=history&includeAdjustedClose=true"
            )
            .dropna()
            .set_index("Date")
        )


        if quotes is not None and len(quotes) > 1:
            quotes["date64"] = pd.to_datetime(quotes.index, format="%Y-%m-%d")
            quotes = quotes.reset_index(drop=True).set_index("date64").sort_index()
            quotes.index.name = "date"
            quotes.columns = [
                "open",
                "high",
                "low",
                "close",
                "adjusted_close",
                "volume",
            ]

            dividends = cls._download_dividends_yahoo(
                signals_ticker, date_from, date_until
            )
            splits = cls._download_splits_yahoo(
                signals_ticker, date_from, date_until
            )

            if dividends is not None and len(dividends) >= 1:
                quotes = quotes.join(dividends, how="left")


            if splits is not None and len(splits) >= 1:
                quotes = quotes.join(splits, how="left")

        return quotes


class EODHDDownloaderOHLCV:
    def __init__(self):
        pass

    @staticmethod
    def load_dividends_eodhd(ticker, api_key, date_from):
        """
        Load the splits data from the EOD Historical Data API.
        """
        url = f"https://eodhistoricaldata.com/api/div/{ticker}?api_token={api_key}&fmt=json&from={date_from}"
        response = requests.get(url)

        if response.status_code == 200:
            if len(response.json()) > 0:
                res = (
                    pd.DataFrame(response.json())
                    .set_index("date")
                    .add_prefix("dividend_")
                )
                res.index = pd.to_datetime(res.index, format="%Y-%m-%d")
                res.rename(columns={"dividend_value": "dividend_amount"}, inplace=True)
                # keep only the dividend amount and date
                res = res[["dividend_amount"]]
                return res
        else:
            return None

    @staticmethod
    def load_splits_eodhd(ticker, api_key, date_from):
        """
        Load the splits data from the EOD Historical Data API.
        """
        url = f"https://eodhistoricaldata.com/api/splits/{ticker}?api_token={api_key}&fmt=json&from={date_from}"
        response = requests.get(url)

        if response.status_code == 200:
            if len(response.json()) > 0:
                df = (
                    pd.DataFrame(response.json())
                    .set_index("date")
                    .rename(columns={"split": "split_ratio"})
                )
                # parse the split ratio from string to float: '2.000000/1.000000' -> 2.0
                df["split_ratio"] = df["split_ratio"].apply(
                    lambda x: float(x.split("/")[0])
                )
                df.index = pd.to_datetime(df.index, format="%Y-%m-%d")
                return df
        else:
            return None

    @classmethod
    def eodhd_download_one(
        cls, signals_ticker, api_key, date_from=None, date_until=None
    ):
        if date_from is None:
            start_date = "2000-01-01"
        else:
            start_date = date_from

        quotes = None

        r = requests.get(
            f"https://eodhistoricaldata.com/api/eod/{signals_ticker}?from={start_date}&fmt=json&api_token={api_key}"
        )

        if r.status_code == requests.codes.ok:
            if len(r.json()) > 0:
                quotes = pd.DataFrame(r.json()).set_index("date")
                quotes["date64"] = pd.to_datetime(quotes.index, format="%Y-%m-%d")
                quotes = quotes.reset_index(drop=True).set_index("date64").sort_index()
                quotes.index.name = "date"
                quotes.columns = [
                    "open",
                    "high",
                    "low",
                    "close",
                    "adjusted_close",
                    "volume",
                ]

            dividends = cls.load_dividends_eodhd(signals_ticker, api_key, date_from)
            splits = cls.load_splits_eodhd(signals_ticker, api_key, date_from)

            if dividends is not None:
                quotes: pd.DataFrame = quotes.join(dividends, how="left", on="date")

            if splits is not None:
                quotes: pd.DataFrame = quotes.join(splits, how="left", on="date")

        return quotes


class StockDataDownloader:
    def __init__(self, max_workers=2, eodhd_apikey: str = ""):
        self.max_workers = max_workers
        self.eodhd_apikey = eodhd_apikey

    def download_one(self, bloomberg_ticker, map, eodhd_api_key=None, date_from=None):
        yahoo_ticker = map.loc[bloomberg_ticker, "yahoo"]
        signals_ticker = map.loc[bloomberg_ticker, "signals_ticker"]
        data_provider = map.loc[bloomberg_ticker, "data_provider"]

        if self.eodhd_apikey is None or len(self.eodhd_apikey) == 0:
            data_provider = "yahoo"

        if pd.isnull(signals_ticker):
            return bloomberg_ticker, None

        quotes = None
        for _ in range(3):
            try:
                if data_provider == "eodhd":
                    quotes = EODHDDownloaderOHLCV.eodhd_download_one(
                        signals_ticker, eodhd_api_key, date_from=date_from
                    )
                elif data_provider == "yahoo":
                    quotes = YahooDownloaderOHLCV.yahoo_download_one(
                        signals_ticker=signals_ticker, date_from=date_from
                    )

                if quotes is not None:
                    quotes["data_provider"] = data_provider

                break

            except Exception as ex:
                # logger.exception(ex)
                time.sleep(5)

        return bloomberg_ticker, quotes

    def download_all(self, ticker_map, date_from=None):
        tickers = pd.Series(ticker_map.index).sample(frac=1).unique().tolist()
        print(f"download_all, tickers:{len(tickers)}")

        all_quotes = []
        eodhd_api_key = self.eodhd_apikey

        with futures.ThreadPoolExecutor(self.max_workers) as executor:
            _futures = []
            for ticker in tqdm(tickers):
                _futures.append(
                    executor.submit(
                        self.download_one,
                        bloomberg_ticker=ticker,
                        map=ticker_map,
                        eodhd_api_key=eodhd_api_key,
                        date_from=date_from,
                    )
                )

            print(f"download_all, futures:{len(_futures)}")
            for future in tqdm(futures.as_completed(_futures), total=len(tickers)):
                bloomberg_ticker, quotes = future.result()
                if quotes is not None:
                    quotes["bloomberg_ticker"] = bloomberg_ticker
                    all_quotes.append(quotes)

        return all_quotes


def remove_wrong_rows(df: pd.DataFrame) -> pd.DataFrame:
    df = df[df["open"] > 0]
    df = df[df["high"] > 0]
    df = df[df["low"] > 0]
    df = df[df["close"] > 0]
    df = df[df["adjusted_close"] > 0]
    df = df[df["volume"] > 0]
    return df


def re_adjust_ohlc(df: pd.DataFrame) -> pd.DataFrame:
    ratio = df["close"] / df["adjusted_close"]
    df["open"] = df["open"] / ratio
    df["high"] = df["high"] / ratio
    df["low"] = df["low"] / ratio
    df["close"] = df["close"] / ratio
    return df


def update_daily_data(data_dir: str, daily_data_dir: str, EODHD_API_KEY: str = None):
    # read the latest date
    latest_date = get_latest_date(daily_data_dir)

    today_date = datetime.today().strftime("%Y-%m-%d")
    print(f"Today date: {today_date}")
    if today_date == latest_date:
        print("Already up-to-date")
        return
    print(f"Latest date: {latest_date}")

    # download data from the latest date
    downloader = StockDataDownloader(
        max_workers=multiprocessing.cpu_count() - 1, eodhd_apikey=EODHD_API_KEY
    )

    ticker_map_fname = f"{data_dir}/eodhd-map.csv"
    if not os.path.exists(ticker_map_fname):
        print(f"Missing ticker map file: {ticker_map_fname}")
        url = "https://raw.githubusercontent.com/parmarsuraj99/dsignals/main/db/eodhd-map.csv"
        r = requests.get(url)
        if r.status_code == requests.codes.ok:
            with open(ticker_map_fname, "wb") as f:
                f.write(r.content)

    ticker_map = pd.read_csv(f"{data_dir}/eodhd-map.csv").set_index("bloomberg_ticker")

    all_quotes = downloader.download_all(ticker_map, date_from=latest_date)

    # save all quotes
    all_quotes = pd.concat(all_quotes)
    all_quotes = remove_wrong_rows(all_quotes)
    all_quotes = re_adjust_ohlc(all_quotes)

    print(all_quotes.index)
    print(all_quotes.columns)

    print(all_quotes.shape)
    print(all_quotes.head())

    save_in_folders(all_quotes, daily_data_dir)

    # if os.path.exists(f"{data_dir}/all_quotes.parquet"):
    #     _prev_quotes = pd.read_pickle(f"{data_dir}/all_quotes.parquet")
    #     # concat the new quotes with the old ones
    #     all_quotes = pd.concat([_prev_quotes, all_quotes], axis=0)

    # try:
    #     all_quotes.to_parquet(f"{data_dir}/all_quotes.parquet")
    # except Exception as ex:
    #     print(f"Failed to save all quotes: {ex}")
    gc.collect()


if __name__ == "__main__":
    dir_config = Directories()
    update_daily_data(dir_config.DATA_DIR, dir_config.DAILY_DATA_DIR)
    pass



import os
from pathlib import Path
import pandas as pd
import gc
import multiprocessing
from joblib import Parallel, delayed
from tqdm import tqdm
from functools import lru_cache


def read_available_dates(DAILY_DATA_DIR):
    dates = []
    for root, dirs, files in os.walk(DAILY_DATA_DIR):
        for file in files:
            if file.endswith(".parquet"):
                dates.append(file[:-8])
    dates = sorted(dates)
    return dates


@lru_cache(maxsize=256)
def load_recent_data_from_file(
    DAILY_DATA_DIR, n_days=1, ascending=False, offset=0, dtype: str = "float32"
):
    assert dtype in ["float32", "float16"]
    print(DAILY_DATA_DIR)

    dates = []
    for root, dirs, files in os.walk(DAILY_DATA_DIR):
        for file in files:
            if file.endswith(".parquet"):
                dates.append(file[:-8])
    _to_reverse = ascending == False
    dates = sorted(dates, reverse=_to_reverse)

    if n_days < 0:
        _tmp = pd.concat(
            [
                pd.read_parquet(
                    os.path.join(DAILY_DATA_DIR, date[:4], date[5:7], f"{date}.parquet")
                )
                for date in dates
            ]
        )
        if dtype == "float16":
            float16_cols = _tmp.select_dtypes(include=["float32"]).columns
            _tmp[float16_cols] = _tmp[float16_cols].astype("float16")

        return _tmp
    
    dates = dates[offset: offset + n_days]

    _tmp = pd.concat(
        [
            pd.read_parquet(
                os.path.join(DAILY_DATA_DIR, date[:4], date[5:7], f"{date}.parquet")
            )
            for date in dates[-n_days:]
        ]
    )
    if dtype == "float16":
        float16_cols = _tmp.select_dtypes(include=["float32"]).columns
        _tmp[float16_cols] = _tmp[float16_cols].astype("float16")

    return _tmp


def save_daily_data(df, date, level_path):
    year = date[:4]
    month = date[5:7]
    day = date[8:10]
    Path(os.path.join(level_path, year, month)).mkdir(parents=True, exist_ok=True)
    f_name = os.path.join(level_path, year, month, f"{date}.parquet")
    # print(f_name)
    if not os.path.exists(f_name):
        # print(f"Saving {date} data to {f_name}")
        df.to_parquet(f_name)


def save_in_folders(df, level_path):
    if "date_str" not in df.columns and "date" not in df.columns:
        df.index = pd.to_datetime(df.index)
        df.index.name = "date"
        df["date_str"] = df.index.strftime("%Y-%m-%d")

    if "date_str" not in df.columns:
        df["date"] = pd.to_datetime(df["date"])
        df["date_str"] = df["date"].dt.strftime("%Y-%m-%d")

    Path(os.path.join(level_path)).mkdir(parents=True, exist_ok=True)

    # save data for each date in a separate file with parquet format
    res = Parallel(n_jobs=multiprocessing.cpu_count())(
        delayed(save_daily_data)(group, date, level_path)
        for date, group in tqdm(df.groupby("date_str"))
    )
    del res

    gc.collect()


def get_latest_date(DAILY_DATA_DIR: str):
    dates = []
    for root, dirs, files in os.walk(DAILY_DATA_DIR):
        for file in files:
            if file.endswith(".parquet"):
                dates.append(file[:-8])
    return max(dates) if len(dates) > 0 else "2000-01-01"



import numpy as np
import pandas as pd
import os
from pathlib import Path
import gc
import multiprocessing
from joblib import Parallel, delayed
from tqdm import tqdm

from typing import Union, List, Tuple, Optional, Dict, Any

import itertools

from signalslite.data_utils import (
    load_recent_data_from_file,
    save_daily_data,
    save_in_folders,
    get_latest_date,
    read_available_dates,
)
from signalslite.constants import Directories

try:
    import cudf
    import numba
    from numba import cuda
except ImportError:
    print("cudf not found, using pandas")


def get_combination_ratio(df, feature_prefix: str):
    # fetch feature_name and its combination ratio

    feature_cols = [f for f in df.columns if feature_prefix in f]
    feature_pairs = itertools.combinations(feature_cols, 2)

    _feature_list = []
    for _f1, _f2 in feature_pairs:
        _feature_type = _f1.split("_")[2]  # sma, ema, macd, rsi, etc.
        _f1_window_size = _f1.split("_")[-1]  # window size 1
        _f2_window_size = _f2.split("_")[-1]  # window size 2

        _res = 1 - (df[_f2] / df[_f1])
        _res.name = (
            f"feature_2_ratio_{_feature_type}_{_f1_window_size}_{_f2_window_size}"
        )

        _feature_list.append(_res)

        gc.collect()

    # ratio with close price for each feature
    for _f in feature_cols:
        _feature_type = _f.split("_")[2]
        _f_window_size = _f.split("_")[-1]

        _res = 1 - (df["close"] / df[_f])
        _res.name = f"feature_2_ratio_{_feature_type}_{_f_window_size}_close"

        _feature_list.append(_res)

        gc.collect()

    _cated_res = pd.concat(_feature_list, axis=1).astype("float32")

    del _feature_list
    gc.collect()

    return _cated_res


def calculate_all_secondary_features(
    df, feature_prefixes: List[str]
):
    _all_features = []
    for feature_prefix in feature_prefixes:
        _features = [f for f in df.columns if feature_prefix in f]
        _res = get_combination_ratio(df.loc[:, _features + ["close"]], feature_prefix)
        _all_features.append(_res)
        gc.collect()

    _all_features = pd.concat(_all_features, axis=1).astype("float32")
    return _all_features


def update_secondary_features(dir_config, feature_prefixes=None):
    if feature_prefixes is None:
        feature_prefixes = [
            "feature_1_sma",
            "feature_1_ema",
            "feature_1_rsi",
        ]
    start_index = 0
    primary_data_dates = read_available_dates(dir_config.DAILY_PRIMARY_FEATURES_DIR)
    print(f"primary_dates: {len(primary_data_dates)}")
    dates = primary_data_dates

    if os.path.exists(dir_config.DAILY_SECONDARY_FEATURES_DIR):
        secondary_data_dates = read_available_dates(
            dir_config.DAILY_SECONDARY_FEATURES_DIR
        )
        print(f"secondary_features_dates: {len(secondary_data_dates)}")

        n_days_to_load = len(primary_data_dates) - len(secondary_data_dates) + 10

        print(f"n_days_to_load: {n_days_to_load}")

        dates = primary_data_dates
        start_index = max(
            0, len(secondary_data_dates) - 1000
        )  # to avoid getting negative dates

    # iterate over all dates in chunks of 200
    for i in tqdm(range(start_index, len(dates), 1000)):
        print(i)
        _df = load_recent_data_from_file(
            dir_config.DAILY_PRIMARY_FEATURES_DIR,
            n_days=1000,
            ascending=True,
            offset=i,
            dtype="float32",
        )

        feat_cols = [f for f in _df if "feature_" in f]

        _res = calculate_all_secondary_features(_df, feature_prefixes).astype("float32")

        # combine primary and secondary featuers
        _res = pd.concat([_df, _res], axis=1)
        _res = _res.replace([np.inf, -np.inf], np.nan)
        _res = _res.dropna(axis=0)

        assert (
            _res.isna().mean().sort_values(ascending=False).max() < 0.1
        ), "too many NaN values found"
        save_in_folders(_res, dir_config.DAILY_SECONDARY_FEATURES_DIR)

        # del _df, _res
        gc.collect()


if __name__ == "__main__":
    dir_config = Directories()

    feature_prefixes = [
        "feature_1_sma",
        "feature_1_ema",
        "feature_1_rsi",
    ]

    update_secondary_features(dir_config, feature_prefixes)
